@@ -52,12 +52,14 @@ macro_rules! try_opt_or {
     };
 }
 
-pub mod parser;
+mod parser;
 mod tables;
+#[cfg(feature = "apple-layout")] mod aat;
 #[cfg(feature = "opentype-layout")] mod ggg;
 #[cfg(feature = "variable-fonts")] mod var_store;
 
-use parser::{Stream, FromData, NumFrom, TryNumFrom, LazyArray16, LazyArrayIter16, Offset32, Offset};
+use parser::{Stream, NumFrom, TryNumFrom, Offset32, Offset};
+pub use parser::{FromData, LazyArray16, LazyArrayIter16, LazyArray32, LazyArrayIter32, Fixed};
 use head::IndexToLocationFormat;
 
 #[cfg(feature = "variable-fonts")] pub use fvar::VariationAxis;
@@ -68,16 +70,26 @@ pub use tables::CFFError;
 pub use tables::{cmap, kern, sbix, maxp, hmtx, name, os2, loca, svg, vorg, post, head, hhea, glyf};
 pub use tables::{cff1 as cff, vhea, cbdt, cblc};
 #[cfg(feature = "opentype-layout")] pub use tables::{gdef, gpos, gsub};
+#[cfg(feature = "apple-layout")] pub use tables::{ankr, feat, kerx, morx, trak};
 #[cfg(feature = "variable-fonts")] pub use tables::{cff2, avar, fvar, gvar, hvar, mvar};
 
 #[cfg(feature = "opentype-layout")]
 pub mod opentype_layout {
     //! This module contains
     //! [OpenType Layout](https://docs.microsoft.com/en-us/typography/opentype/spec/chapter2#overview)
-    //! tables implementation.
+    //! supplementary tables implementation.
     pub use crate::ggg::*;
 }
 
+#[cfg(feature = "apple-layout")]
+pub mod apple_layout {
+    //! This module contains
+    //! [Apple Advanced Typography Layout](
+    //! https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6AATIntro.html)
+    //! supplementary tables implementation.
+    pub use crate::aat::*;
+}
+
 
 /// A type-safe wrapper for glyph ID.
 #[repr(transparent)]
@@ -183,6 +195,12 @@ pub struct Tag(pub u32);
 
 impl Tag {
     /// Creates a `Tag` from bytes.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// println!("{}", ttf_parser::Tag::from_bytes(b"name"));
+    /// ```
     #[inline]
     pub const fn from_bytes(bytes: &[u8; 4]) -> Self {
         Tag(((bytes[0] as u32) << 24) | ((bytes[1] as u32) << 16) |
@@ -531,6 +549,85 @@ impl core::fmt::Display for FaceParsingError {
 impl std::error::Error for FaceParsingError {}
 
 
+/// A raw font face.
+///
+/// You are probably looking for [`Face`]. This is a low-level type.
+///
+/// Unlike [`Face`], [`RawFace`] parses only face table records.
+/// Meaning all you can get from this type is a raw (`&[u8]`) data of a requested table.
+/// Then you can either parse just a singe table from a font/face or populate [`RawFaceTables`]
+/// manually before passing it to [`Face::from_raw_tables`].
+#[derive(Clone, Copy)]
+pub struct RawFace<'a> {
+    data: &'a [u8],
+    table_records: LazyArray16<'a, TableRecord>,
+}
+
+impl<'a> RawFace<'a> {
+    /// Creates a new [`RawFace`] from a raw data.
+    ///
+    /// `index` indicates the specific font face in a font collection.
+    /// Use [`fonts_in_collection`] to get the total number of font faces.
+    /// Set to 0 if unsure.
+    ///
+    /// While we do reuse [`FaceParsingError`], `No*Table` errors will not be throws.
+    pub fn from_slice(data: &'a [u8], index: u32) -> Result<Self, FaceParsingError> {
+        // https://docs.microsoft.com/en-us/typography/opentype/spec/otff#organization-of-an-opentype-font
+
+        let mut s = Stream::new(data);
+
+        // Read **font** magic.
+        let magic = s.read::<Magic>().ok_or(FaceParsingError::UnknownMagic)?;
+        if magic == Magic::FontCollection {
+            s.skip::<u32>(); // version
+            let number_of_faces = s.read::<u32>().ok_or(FaceParsingError::MalformedFont)?;
+            let offsets = s.read_array32::<Offset32>(number_of_faces)
+                .ok_or(FaceParsingError::MalformedFont)?;
+
+            let face_offset = offsets.get(index).ok_or(FaceParsingError::FaceIndexOutOfBounds)?;
+            // Face offset is from the start of the font data,
+            // so we have to adjust it to the current parser offset.
+            let face_offset = face_offset.to_usize().checked_sub(s.offset())
+                .ok_or(FaceParsingError::MalformedFont)?;
+            s.advance_checked(face_offset).ok_or(FaceParsingError::MalformedFont)?;
+
+            // Read **face** magic.
+            // Each face in a font collection also starts with a magic.
+            let magic = s.read::<Magic>().ok_or(FaceParsingError::UnknownMagic)?;
+            // And face in a font collection can't be another collection.
+            if magic == Magic::FontCollection {
+                return Err(FaceParsingError::UnknownMagic);
+            }
+        }
+
+        let num_tables = s.read::<u16>().ok_or(FaceParsingError::MalformedFont)?;
+        s.advance(6); // searchRange (u16) + entrySelector (u16) + rangeShift (u16)
+        let table_records = s.read_array16::<TableRecord>(num_tables)
+            .ok_or(FaceParsingError::MalformedFont)?;
+
+        Ok(RawFace {
+            data,
+            table_records,
+        })
+    }
+
+    /// Returns the raw data of a selected table.
+    pub fn table(&self, tag: Tag) -> Option<&'a [u8]> {
+        let (_, table) = self.table_records.binary_search_by(|record| record.table_tag.cmp(&tag))?;
+        let offset = usize::num_from(table.offset);
+        let length = usize::num_from(table.length);
+        let end = offset.checked_add(length)?;
+        self.data.get(offset..end)
+    }
+}
+
+impl core::fmt::Debug for RawFace<'_> {
+    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
+        write!(f, "RawFace {{ ... }}")
+    }
+}
+
+
 /// A list of all supported tables as raw data.
 ///
 /// This type should be used in tandem with
@@ -568,6 +665,12 @@ pub struct RawFaceTables<'a> {
     #[cfg(feature = "opentype-layout")] pub gpos: Option<&'a [u8]>,
     #[cfg(feature = "opentype-layout")] pub gsub: Option<&'a [u8]>,
 
+    #[cfg(feature = "apple-layout")] pub ankr: Option<&'a [u8]>,
+    #[cfg(feature = "apple-layout")] pub feat: Option<&'a [u8]>,
+    #[cfg(feature = "apple-layout")] pub kerx: Option<&'a [u8]>,
+    #[cfg(feature = "apple-layout")] pub morx: Option<&'a [u8]>,
+    #[cfg(feature = "apple-layout")] pub trak: Option<&'a [u8]>,
+
     #[cfg(feature = "variable-fonts")] pub avar: Option<&'a [u8]>,
     #[cfg(feature = "variable-fonts")] pub cff2: Option<&'a [u8]>,
     #[cfg(feature = "variable-fonts")] pub fvar: Option<&'a [u8]>,
@@ -612,6 +715,12 @@ pub struct FaceTables<'a> {
     #[cfg(feature = "opentype-layout")] pub gpos: Option<opentype_layout::LayoutTable<'a>>,
     #[cfg(feature = "opentype-layout")] pub gsub: Option<opentype_layout::LayoutTable<'a>>,
 
+    #[cfg(feature = "apple-layout")] pub ankr: Option<ankr::Table<'a>>,
+    #[cfg(feature = "apple-layout")] pub feat: Option<feat::Table<'a>>,
+    #[cfg(feature = "apple-layout")] pub kerx: Option<kerx::Table<'a>>,
+    #[cfg(feature = "apple-layout")] pub morx: Option<morx::Table<'a>>,
+    #[cfg(feature = "apple-layout")] pub trak: Option<trak::Table<'a>>,
+
     #[cfg(feature = "variable-fonts")] pub avar: Option<avar::Table<'a>>,
     #[cfg(feature = "variable-fonts")] pub cff2: Option<cff2::Table<'a>>,
     #[cfg(feature = "variable-fonts")] pub fvar: Option<fvar::Table<'a>>,
@@ -638,8 +747,7 @@ pub struct FaceTables<'a> {
 /// While `Face` is technically copyable, we disallow it because it's almost 2KB big.
 #[derive(Clone)]
 pub struct Face<'a> {
-    font_data: &'a [u8], // The input data. Used by Face::table_data.
-    table_records: LazyArray16<'a, TableRecord>, // Raw table data.
+    raw_face: RawFace<'a>,
     tables: FaceTables<'a>, // Parsed tables.
     #[cfg(feature = "variable-fonts")] coordinates: VarCoords,
 }
@@ -648,7 +756,7 @@ impl<'a> Face<'a> {
     /// Creates a new [`Face`] from a raw data.
     ///
     /// `index` indicates the specific font face in a font collection.
-    /// Use `fonts_in_collection` to get the total number of font faces.
+    /// Use [`fonts_in_collection`] to get the total number of font faces.
     /// Set to 0 if unsure.
     ///
     /// This method will do some parsing and sanitization,
@@ -658,45 +766,12 @@ impl<'a> Face<'a> {
     ///
     /// If an optional table has invalid data it will be skipped.
     pub fn from_slice(data: &'a [u8], index: u32) -> Result<Self, FaceParsingError> {
-        // https://docs.microsoft.com/en-us/typography/opentype/spec/otff#organization-of-an-opentype-font
-
-        let mut s = Stream::new(data);
-
-        // Read **font** magic.
-        let magic = s.read::<Magic>().ok_or(FaceParsingError::UnknownMagic)?;
-        if magic == Magic::FontCollection {
-            s.skip::<u32>(); // version
-            let number_of_faces = s.read::<u32>().ok_or(FaceParsingError::MalformedFont)?;
-            let offsets = s.read_array32::<Offset32>(number_of_faces)
-                .ok_or(FaceParsingError::MalformedFont)?;
-
-            let face_offset = offsets.get(index).ok_or(FaceParsingError::FaceIndexOutOfBounds)?;
-            // Face offset is from the start of the font data,
-            // so we have to adjust it to the current parser offset.
-            let face_offset = face_offset.to_usize().checked_sub(s.offset())
-                .ok_or(FaceParsingError::MalformedFont)?;
-            s.advance_checked(face_offset).ok_or(FaceParsingError::MalformedFont)?;
-
-            // Read **face** magic.
-            // Each face in a font collection also starts with a magic.
-            let magic = s.read::<Magic>().ok_or(FaceParsingError::UnknownMagic)?;
-            // And face in a font collection can't be another collection.
-            if magic == Magic::FontCollection {
-                return Err(FaceParsingError::UnknownMagic);
-            }
-        }
-
-        let num_tables = s.read::<u16>().ok_or(FaceParsingError::MalformedFont)?;
-        s.advance(6); // searchRange (u16) + entrySelector (u16) + rangeShift (u16)
-        let table_records = s.read_array16::<TableRecord>(num_tables)
-            .ok_or(FaceParsingError::MalformedFont)?;
-
-        let raw_tables = Self::collect_tables(data, table_records);
+        let raw_face = RawFace::from_slice(data, index)?;
+        let raw_tables = Self::collect_tables(raw_face);
 
         #[allow(unused_mut)]
         let mut face = Face {
-            font_data: data,
-            table_records: table_records,
+            raw_face,
             #[cfg(feature = "variable-fonts")] coordinates: VarCoords::default(),
             tables: Self::parse_tables(raw_tables)?,
         };
@@ -710,20 +785,17 @@ impl<'a> Face<'a> {
         Ok(face)
     }
 
-    fn collect_tables(
-        data: &'a [u8],
-        records: LazyArray16<TableRecord>,
-    ) -> RawFaceTables<'a> {
+    fn collect_tables(raw_face: RawFace<'a>) -> RawFaceTables<'a> {
         let mut tables = RawFaceTables::default();
 
-        for record in records {
+        for record in raw_face.table_records {
             let start = usize::num_from(record.offset);
             let end = match start.checked_add(usize::num_from(record.length)) {
                 Some(v) => v,
                 None => continue,
             };
 
-            let table_data = data.get(start..end);
+            let table_data = raw_face.data.get(start..end);
             match &record.table_tag.to_bytes() {
                 b"CBDT" => tables.cbdt = table_data,
                 b"CBLC" => tables.cblc = table_data,
@@ -745,9 +817,13 @@ impl<'a> Face<'a> {
                 b"VORG" => tables.vorg = table_data,
                 #[cfg(feature = "variable-fonts")]
                 b"VVAR" => tables.vvar = table_data,
+                #[cfg(feature = "apple-layout")]
+                b"ankr" => tables.ankr = table_data,
                 #[cfg(feature = "variable-fonts")]
                 b"avar" => tables.avar = table_data,
                 b"cmap" => tables.cmap = table_data,
+                #[cfg(feature = "apple-layout")]
+                b"feat" => tables.feat = table_data,
                 #[cfg(feature = "variable-fonts")]
                 b"fvar" => tables.fvar = table_data,
                 b"glyf" => tables.glyf = table_data,
@@ -757,11 +833,17 @@ impl<'a> Face<'a> {
                 b"hhea" => tables.hhea = table_data.unwrap_or_default(),
                 b"hmtx" => tables.hmtx = table_data,
                 b"kern" => tables.kern = table_data,
+                #[cfg(feature = "apple-layout")]
+                b"kerx" => tables.kerx = table_data,
                 b"loca" => tables.loca = table_data,
                 b"maxp" => tables.maxp = table_data.unwrap_or_default(),
+                #[cfg(feature = "apple-layout")]
+                b"morx" => tables.morx = table_data,
                 b"name" => tables.name = table_data,
                 b"post" => tables.post = table_data,
                 b"sbix" => tables.sbix = table_data,
+                #[cfg(feature = "apple-layout")]
+                b"trak" => tables.trak = table_data,
                 b"vhea" => tables.vhea = table_data,
                 b"vmtx" => tables.vmtx = table_data,
                 _ => {}
@@ -775,8 +857,10 @@ impl<'a> Face<'a> {
     pub fn from_raw_tables(raw_tables: RawFaceTables<'a>) -> Result<Self, FaceParsingError> {
         #[allow(unused_mut)]
         let mut face = Face {
-            font_data: &[],
-            table_records: LazyArray16::default(),
+            raw_face: RawFace {
+                data: &[],
+                table_records: LazyArray16::default(),
+            },
             #[cfg(feature = "variable-fonts")] coordinates: VarCoords::default(),
             tables: Self::parse_tables(raw_tables)?,
         };
@@ -847,6 +931,15 @@ impl<'a> Face<'a> {
             #[cfg(feature = "opentype-layout")] gpos: raw_tables.gpos.and_then(opentype_layout::LayoutTable::parse),
             #[cfg(feature = "opentype-layout")] gsub: raw_tables.gsub.and_then(opentype_layout::LayoutTable::parse),
 
+            #[cfg(feature = "apple-layout")] ankr: raw_tables.ankr
+                .and_then(|data| ankr::Table::parse(maxp.number_of_glyphs, data)),
+            #[cfg(feature = "apple-layout")] feat: raw_tables.feat.and_then(feat::Table::parse),
+            #[cfg(feature = "apple-layout")] kerx: raw_tables.kerx
+                .and_then(|data| kerx::Table::parse(maxp.number_of_glyphs, data)),
+            #[cfg(feature = "apple-layout")] morx: raw_tables.morx
+                .and_then(|data| morx::Table::parse(maxp.number_of_glyphs, data)),
+            #[cfg(feature = "apple-layout")] trak: raw_tables.trak.and_then(trak::Table::parse),
+
             #[cfg(feature = "variable-fonts")] avar: raw_tables.avar.and_then(avar::Table::parse),
             #[cfg(feature = "variable-fonts")] cff2: raw_tables.cff2.and_then(cff2::Table::parse),
             #[cfg(feature = "variable-fonts")] fvar: raw_tables.fvar.and_then(fvar::Table::parse),
@@ -868,12 +961,9 @@ impl<'a> Face<'a> {
     /// Useful if you want to parse the data manually.
     ///
     /// Available only for faces created using [`Face::from_slice()`](struct.Face.html#method.from_slice).
+    #[inline]
     pub fn table_data(&self, tag: Tag) -> Option<&'a [u8]> {
-        let (_, table) = self.table_records.binary_search_by(|record| record.table_tag.cmp(&tag))?;
-        let offset = usize::num_from(table.offset);
-        let length = usize::num_from(table.length);
-        let end = offset.checked_add(length)?;
-        self.font_data.get(offset..end)
+        self.raw_face.table(tag)
     }
 
     /// Returns a list of names.
